@@ -8,13 +8,14 @@ from typing import (
     List,
     MutableSequence,
     Optional,
+    Sequence,
     Set,
     Tuple,
 )
 
 from transformers import PreTrainedTokenizer
 
-from pytorch_ie.data import LabeledSpan
+from pytorch_ie import LabeledSpan, Span
 
 TypedSpan = Tuple[int, Tuple[int, int]]
 TypedStringSpan = Tuple[str, Tuple[int, int]]
@@ -32,7 +33,9 @@ class InvalidTagSequence(Exception):
 
 
 def bio_tags_to_spans(
-    tag_sequence: List[str], classes_to_ignore: List[str] = None, include_ill_formed: bool = True
+    tag_sequence: Sequence[str],
+    classes_to_ignore: List[str] = None,
+    include_ill_formed: bool = True,
 ) -> List[TypedStringSpan]:
     """
     Given a sequence corresponding to BIO tags, extracts spans.
@@ -157,10 +160,10 @@ def io_tags_to_spans(
 
 
 def convert_span_annotations_to_tag_sequence(
-    spans: List[LabeledSpan],
-    special_tokens_mask: List[int],
+    spans: Sequence[LabeledSpan],
+    special_tokens_mask: Sequence[int],
     char_to_token_mapper: Callable[[int], Optional[int]],
-    partition: Optional[LabeledSpan] = None,
+    partition: Optional[Span] = None,
     statistics: Optional[DefaultDict[str, Counter]] = None,
 ) -> MutableSequence[Optional[str]]:
     """
@@ -183,7 +186,7 @@ def convert_span_annotations_to_tag_sequence(
         end_idx = char_to_token_mapper(span.end - 1 - offset)
         if start_idx is None or end_idx is None:
             if statistics is not None:
-                statistics["skipped_unaligned"][span.label_single] += 1
+                statistics["skipped_unaligned"][span.label] += 1
             else:
                 logger.warning(
                     f"Entity annotation does not start or end with a token, it will be skipped: {span}"
@@ -199,10 +202,10 @@ def convert_span_annotations_to_tag_sequence(
                 # TODO: is ValueError a good exception type for this?
                 raise ValueError(f"tag already assigned (current span has an overlap: {span})")
             prefix = "B" if j == start_idx else "I"
-            tag_sequence[j] = f"{prefix}-{span.label_single}"
+            tag_sequence[j] = f"{prefix}-{span.label}"
 
         if statistics is not None:
-            statistics["added"][span.label_single] += 1
+            statistics["added"][span.label] += 1
 
     return tag_sequence
 
